import numpy as np
import os
import time
import random
import logging
from tqdm import tqdm
import torch
from torch.utils.data import Dataset, DataLoader
import monai
from segment_anything import sam_model_registry
from segment_anything.utils.transforms import ResizeLongestSide
from pynvml import nvmlInit, nvmlDeviceGetHandleByIndex, nvmlDeviceGetMemoryInfo, nvmlDeviceGetPowerUsage, nvmlShutdown

torch.manual_seed(2024)

# 设置日志
logging.basicConfig(filename="training_log.log", level=logging.INFO, format="%(asctime)s - %(message)s")

# 初始化 GPU 显存监控
nvmlInit()
gpu_handle = nvmlDeviceGetHandleByIndex(0)  # 使用 GPU 0

# 构造 Dataset
class MyselfDataset(Dataset):
    def __init__(self, data_root):
        print("into init")
        self.data_root = data_root
        self.npz_files = sorted([f for f in os.listdir(self.data_root) if f.endswith('.npz')])
        self.npz_data = [np.load(os.path.join(data_root, f)) for f in self.npz_files]
        self.ori_gts = np.vstack([d['gts'] for d in self.npz_data])
        self.img_embeddings = np.vstack([d['img_embeddings'] for d in self.npz_data])

    def __len__(self):
        return self.ori_gts.shape[0]

    def __getitem__(self, index):
        img_embed = self.img_embeddings[index]
        gt2D = self.ori_gts[index]
        y_indices, x_indices = np.where(gt2D > 0)
        x_min, x_max = np.min(x_indices), np.max(x_indices)
        y_min, y_max = np.min(y_indices), np.max(y_indices)
        H, W = gt2D.shape
        x_min = max(0, x_min - np.random.randint(0, 10))
        x_max = min(W, x_max + np.random.randint(0, 10))
        y_min = max(0, y_min - np.random.randint(0, 10))
        y_max = min(H, y_max + np.random.randint(0, 10))
        bboxes = np.array([x_min, y_min, x_max, y_max])

        y_zero, x_zero = np.where(gt2D == 0)
        y_zero = np.unique(y_zero)
        x_zero = np.unique(x_zero)
        y_list = y_zero[(y_min - 5 < y_zero) & (y_zero < y_max + 5)]
        x_list = x_zero[(x_min - 5 < x_zero) & (x_zero < x_max + 5)]
        y1, y2 = random.choices(y_list, k=2)
        x1, x2 = random.choices(x_list, k=2)
        background_index1 = [x1, y1]
        background_index2 = [x2, y2]
        foreground_index1, foreground_index2, foreground_index3 = random.choices(
            np.argwhere(gt2D == 1), k=3)
        points = [background_index1, background_index2, foreground_index1, foreground_index2, foreground_index3]
        points_labels = [0, 0, 1, 1, 1]
        return torch.tensor(img_embed).float(), torch.tensor(gt2D[None, :, :]).long(), torch.tensor(
            bboxes).float(), torch.tensor(points).float(), torch.tensor(points_labels).float()

# 配置训练参数
# 1. 修改npz的存放路径【TODO】
npz_tr_path = '../data/npz/'
# 2. 模型种类【TODO】
model_type = 'vit_h'
# 3. 初始权重文件【TODO】
checkpoint = './sam_vit_h_4b8939.pth'
# 4. 训练迭代步数【TODO】
num_epochs = 200
# 5. 最小loss
best_loss = 0.00001
# 6. 上一次保存的checkpoint的暂存文件名
latest_model_name = "newSAM.pth"
# 7. 最终保存的checkpoint文件名
best_model_name = "bestSAM.pth"
# 8. 指定checkpoint文件存储路径
model_save_path = './'
# 9. 指定主CUDA设备卡号【TODO】
device = 'cuda:0'  # 使用 'cuda:0' 作为主设备
# 10. 指定训练一共调用的卡号【TODO】
os.environ['CUDA_VISIBLE_DEVICES'] = '0,1,2,3'

os.makedirs(model_save_path, exist_ok=True)


# 创建并配置模型
sam_model = sam_model_registry[model_type](checkpoint=checkpoint)
sam_model = sam_model.to(device)
sam_model = torch.nn.DataParallel(sam_model, device_ids=[0])
sam_model.train()

optimizer = torch.optim.Adam(sam_model.module.mask_decoder.parameters(), lr=1e-5, weight_decay=0)
seg_loss_dice = monai.losses.DiceCELoss(sigmoid=True, squared_pred=True, reduction='mean')

losses = []
train_dataset = MyselfDataset(npz_tr_path)
train_dataloader = DataLoader(train_dataset, batch_size=96, shuffle=True)

# 记录数据列表
throughputs, step_times, mem_usages, power_usages, epoch_times = [], [], [], [], []

# 开始训练
for epoch in tqdm(range(num_epochs), desc="Epochs"):
    epoch_loss = 0
    start_time = time.time()
    num_samples = 0

    for step, (image_embedding, gt2D, boxes, points, points_labels) in enumerate(tqdm(train_dataloader)):
        step_start_time = time.time()
        
        with torch.no_grad():
            box_np = boxes.numpy()
            sam_trans = ResizeLongestSide(sam_model.module.image_encoder.img_size)
            box = sam_trans.apply_boxes(box_np, (gt2D.shape[-2], gt2D.shape[-1]))
            box_torch = torch.as_tensor(box, dtype=torch.float, device=device)
            coords_torch = torch.as_tensor(points, dtype=torch.float, device=device)
            labels_torch = torch.as_tensor(points_labels, dtype=torch.int, device=device)
            pt = (coords_torch, labels_torch)
            sparse_embeddings, dense_embeddings = sam_model.module.prompt_encoder(
                points=pt,
                boxes=box_torch,
                masks=None,
            )
        
        mask_predictions, _ = sam_model.module.mask_decoder(
            image_embeddings=image_embedding.to(device),
            image_pe=sam_model.module.prompt_encoder.get_dense_pe(),
            sparse_prompt_embeddings=sparse_embeddings,
            dense_prompt_embeddings=dense_embeddings,
            multimask_output=False,
        )
        
        # 计算损失并优化
        loss_dice = seg_loss_dice(mask_predictions, gt2D.to(device))
        loss = loss_dice
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
        
        # 记录损失
        epoch_loss += loss.item()
        step_time = time.time() - step_start_time
        num_samples += image_embedding.size(0)
        
        throughput = num_samples / (time.time() - start_time)
        mem_info = nvmlDeviceGetMemoryInfo(gpu_handle)
        power_usage = nvmlDeviceGetPowerUsage(gpu_handle) / 1000  # 功率（W）

        throughputs.append(throughput)
        step_times.append(step_time)
        mem_usages.append(mem_info.used / 1024**2)  # MB
        power_usages.append(power_usage)

    epoch_duration = time.time() - start_time
    epoch_times.append(epoch_duration)
    epoch_loss /= step
    losses.append(epoch_loss)
    logging.info(f'Epoch {epoch}, Loss: {epoch_loss:.4f}, Throughput: {throughput:.2f} samples/s')

    # 保存模型
    torch.save(sam_model.module.state_dict(), os.path.join(model_save_path, latest_model_name))
    if epoch_loss < best_loss:
        best_loss = epoch_loss
        torch.save(sam_model.module.state_dict(), os.path.join(model_save_path, best_model_name))

# 关闭 GPU 监控
nvmlShutdown()

# 计算并打印平均指标
avg_throughput = np.mean(throughputs)
avg_step_time = np.mean(step_times)
avg_mem_usage = np.mean(mem_usages)
avg_power_usage = np.mean(power_usages)
avg_epoch_time = np.mean(epoch_times)

logging.info(f"Average Throughput: {avg_throughput:.2f} samples/s")
logging.info(f"Average Step Time: {avg_step_time:.4f}s")
logging.info(f"Average Memory Usage: {avg_mem_usage:.2f} MB")
logging.info(f"Average Power Usage: {avg_power_usage:.2f} W")
logging.info(f"Average Epoch Duration: {avg_epoch_time:.2f}s")

print(f"Average Throughput: {avg_throughput:.2f} samples/s")
print(f"Average Step Time: {avg_step_time:.4f}s")
print(f"Average Memory Usage: {avg_mem_usage:.2f} MB")
print(f"Average Power Usage: {avg_power_usage:.2f} W")
print(f"Average Epoch Duration: {avg_epoch_time:.2f}s")

# 计算 1 万步迭代的成本效益
rent_cost_per_hour = 8  # 每小时 8 元【TODO】
total_steps = 10000
total_time_hours = (total_steps * avg_step_time) / 3600
total_cost = total_time_hours * rent_cost_per_hour

logging.info(f"Cost for 10k steps: {total_cost:.2f} yuan, Total Time: {total_time_hours:.2f} hours")
print(f"Cost for 10k steps: {total_cost:.2f} yuan, Total Time: {total_time_hours:.2f} hours")
